# -*- coding: utf-8 -*-
#
# New BSD license
#
# Copyright (c) DR0ID
# This file is part of pyweek22-2016-09
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL DR0ID BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
TODO: module description


Versioning scheme based on: http://en.wikipedia.org/wiki/Versioning#Designating_development_stage

::

      +-- api change, probably incompatible with older versions
      |     +-- enhancements but no api change
      |     |
    major.minor[.build[.revision]]
                   |        |
                   |        +-|* x for x bugfixes
                   |
                   +-|* 0 for alpha (status)
                     |* 1 for beta (status)
                     |* 2 for release candidate
                     |* 3 for (public) release

.. versionchanged:: 0.0.0.0
    initial version

"""
from __future__ import print_function

import logging
import pygame

logger = logging.getLogger(__name__)

__version__ = '1.0.0.0'

# for easy comparison as in sys.version_info but digits only
__version_info__ = tuple([int(_d) for _d in __version__.split('.')])

__author__ = "DR0ID"
__email__ = "dr0iddr0id {at} gmail [dot] com"
__copyright__ = "DR0ID @ 2016"
__credits__ = ["DR0ID"]  # list of contributors
__maintainer__ = "DR0ID"
__license__ = "New BSD license"

__all__ = []  # list of public visible parts of this module 


def get_actions(event_map, events):
    """
    usage:

    for action, extra in  eventmapper.get_actions(event_map, pygame.event.get()):
        ...

    :param event_map: the event map: {event_type: {(a, b, c): action_x, (a, d, e): action_y}, ...}
    :param events: list of pygame events
    :return: list of actions
    """
    result = []
    for event in events:
        actions = event_map.get(event.type, None)
        if actions is None:
            logger.debug("no actions found for event type: {0}", str(event.type))
            continue
        key = None
        extra = None
        # QUIT             none
        if event.type == pygame.QUIT:
            key = None
        # ACTIVEEVENT      gain, state
        elif event.type == pygame.ACTIVEEVENT:
            key = (event.state, event.gain)
        # KEYDOWN          unicode, key, mod
        elif event.type == pygame.KEYDOWN:
            key = (event.key, event.mod)
        # KEYUP            key, mod
        elif event.type == pygame.KEYUP:
            key = (event.key, event.mod)
        # MOUSEMOTION      pos, rel, buttons
        elif event.type == pygame.MOUSEMOTION:
            key = None
            extra = (event.pos, event.rel, event.buttons)
        # MOUSEBUTTONUP    pos, button
        elif event.type == pygame.MOUSEBUTTONDOWN:
            key = event.button
            extra = (event.pos, )
        # MOUSEBUTTONDOWN  pos, button
        elif event.type == pygame.MOUSEBUTTONUP:
            key = event.button
            extra = (event.pos, )
        # JOYAXISMOTION    joy, axis, value
        # JOYBALLMOTION    joy, ball, rel
        # JOYHATMOTION     joy, hat, value
        # JOYBUTTONUP      joy, button
        # JOYBUTTONDOWN    joy, button
        # VIDEORESIZE      size, w, h
        # VIDEOEXPOSE      none
        # USEREVENT        code
        action = actions.get(key, None)
        if action is not None:
            result.append((action, extra))
        else:
            logger.debug("no mapping found for key: {0}", str(key))

    return result

